#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <EFI.h>

#define MAXLINE 132

typedef unsigned char byte;

extern const char *ProgramName;

void error(char *fmt, ...);
void warning(char *fmt, ...);
void _UART_flush();

struct SRecLine {
  byte type;
  byte count;
  byte DataLength;
  uint address;
  byte checksum;
  byte data[40];
};

byte ReadByte(char **p) {
  register byte a, b;
  register char c;
  
  c=*(*p)++;
  a=(byte)toupper((int)c);
  c=*(*p)++;
  b=(byte)toupper((int)c);
  return ( (((a>0x39)?a-7:a)-0x30)*16+((b>0x39)?b-7:b)-0x30 );
}

byte ReadNib(char **p) {
  register byte a;
  register char c;

  c=*(*p)++;
  a=(byte)toupper((int)c);
  return ( ((a>0x39)?a-7:a)-0x30 );
}

uint ReadAddress(char **p, byte type, byte *NumBytesAddr, byte *checksum) {
  uint addr=0;
  byte a;
  int i;
  
  switch (type) {
    case 0:
    case 1:
    case 9:
      *NumBytesAddr = 2;
      break;
    case 2:
    case 8:
      *NumBytesAddr = 3;
      break;
    case 3:
    case 7:
      *NumBytesAddr = 4;
      break;
    default:
      *NumBytesAddr = 0;
      return 0;
    }
  for (i=0; i<*NumBytesAddr; i++) {
    addr <<= 8;
    a = ReadByte(p);
    *checksum += a;
    addr += a;
  }
  return addr;
}

void SRecLoader(void)
{
  char line[MAXLINE], *p;
  struct SRecLine data;
  int LineNumber = 0;
  int i;
  uint StartAddress=0;
  byte NumBytesAddr;
  byte *pter;
  byte checksum;
        
  ProgramName="SRecLoader";

  printf("\nReady to load program\n");

  while (gets(line) != NULL) {
    while ( isspace(line[(i=strlen(line)-1)]) ) line[i]='\0';
    LineNumber++;
    p=line;
    if (*p++ != 'S') {
      warning("Invalid record at line %i.\n",LineNumber);
      goto IgnoreLine;
    }
    data.type = ReadNib(&p);
    checksum = data.count = ReadByte(&p);
    if (strlen(line) != (data.count+2)*2 ) {
      warning("Line length error at line %i.\n",LineNumber);
      goto IgnoreLine;
    }
    data.address = ReadAddress(&p,data.type,&NumBytesAddr,&checksum);
    if (NumBytesAddr == 0) {
      warning("Unknown 'S' type at line %i.\n",LineNumber);
      goto IgnoreLine;
    }
    data.DataLength = data.count-NumBytesAddr-1;
    for (i = 0; i < data.DataLength; i++)
      checksum += data.data[i] = ReadByte(&p);
    data.checksum = ReadByte(&p);
    checksum = ~checksum;
    if (checksum != data.checksum) {
      warning("Checksum error at line %i.\n",LineNumber);
      goto IgnoreLine;
    }

    /* process the valid data */
    switch (data.type) {
      case 0:
	printf("File name: ");
	for (i=0;i<data.DataLength;i++)
	  putchar(data.data[i]);
	printf("\r\n");
	break;
      case 1:
      case 2:
      case 3:
	printf("Loading line %i.\r",LineNumber);
/* 	printf("Load Adress: %x\n",data.address); */
/* 	for (i=0;i<data.DataLength;i++) */
/* 	  printf("%2x ",data.data[i]); */
/* 	printf("\n"); */
	pter = (byte *)data.address + 0xf0000;
	for (i=0; i<data.DataLength; i++)
	  *pter++ = data.data[i];
	break;
      case 7:
      case 8:
      case 9:
	printf("\nStart Address: %x\n",data.address);
	StartAddress=data.address;
	break;
    }
    IgnoreLine:
  }

  fflush(stdout);
  _UART_flush();

#if 0
  /* start execution */
  if (StartAddress != 0)
    ((void(*)())StartAddress)();
#endif /* DEBUG */

}
