#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#define MAXLINE 132
#define TPU_RAM_SIZE (2*1024)

typedef unsigned char byte;

const char *ProgramName;

byte tpu_buffer[TPU_RAM_SIZE];

struct SRecLine {
  byte type;
  byte count;
  byte DataLength;
  uint address;
  byte checksum;
  byte data[40];
};

void error(char *fmt, ...) {
  va_list args;
  
  va_start(args, fmt);
  fprintf(stderr,"%s: ", ProgramName);
  vfprintf(stderr, fmt, args);
  va_end(args);
  exit(1);
}

void warning(char *fmt, ...) {
  va_list args;
  
  va_start(args, fmt);
  fprintf(stderr,"%s: ", ProgramName);
  vfprintf(stderr, fmt, args);
  va_end(args);
}

byte ReadByte(char **p) {
  register byte a, b;
  register char c;
  
  c=*(*p)++;
  a=(byte)toupper((int)c);
  c=*(*p)++;
  b=(byte)toupper((int)c);
  return ( (((a>0x39)?a-7:a)-0x30)*16+((b>0x39)?b-7:b)-0x30 );
}

byte ReadNib(char **p) {
  register byte a;
  register char c;

  c=*(*p)++;
  a=(byte)toupper((int)c);
  return ( ((a>0x39)?a-7:a)-0x30 );
}

uint ReadAddress(char **p, byte type, byte *NumBytesAddr, byte *checksum) {
  uint addr=0;
  byte a;
  int i;
  
  switch (type) {
    case 0:
    case 1:
    case 9:
      *NumBytesAddr = 2;
      break;
    case 2:
    case 8:
      *NumBytesAddr = 3;
      break;
    case 3:
    case 7:
      *NumBytesAddr = 4;
      break;
    default:
      *NumBytesAddr = 0;
      return 0;
    }
  for (i=0; i<*NumBytesAddr; i++) {
    addr <<= 8;
    a = ReadByte(p);
    *checksum += a;
    addr += a;
  }
  return addr;
}

void proc_data(struct SRecLine data) {
  byte *pter;
  int i;

  switch (data.type) {
  case 0:
    break;
  case 1:
  case 2:
  case 3:
/*     printf("Loading line %i.\r",LineNumber); */
/*     printf("Load Adress: %x\n",data.address); */
/*     for (i=0;i<data.DataLength;i++) */
/*       printf("%2x ",data.data[i]); */
/*     printf("\n"); */
    pter = (byte *)(data.address + (uint)tpu_buffer);
    for (i=0; i<data.DataLength; i++)
      *pter++ = data.data[i];
    break;
  case 7:
  case 8:
  case 9:
    break;
  }
}

void SRecLoader(void)
{
  char line[MAXLINE], *p;
  struct SRecLine data;
  int LineNumber = 0;
  int i;
  byte NumBytesAddr;
  byte checksum;
        
  while (gets(line) != NULL) {
    while ( isspace(line[(i=strlen(line)-1)]) ) line[i]='\0';
    LineNumber++;
    p=line;
    if (*p++ != 'S') {
      warning("Invalid record at line %i.\n",LineNumber);
      continue;
    }
    data.type = ReadNib(&p);
    checksum = data.count = ReadByte(&p);
    if (strlen(line) != (data.count+2)*2 ) {
      warning("Line length error at line %i.\n",LineNumber);
      continue;
    }
    data.address = ReadAddress(&p,data.type,&NumBytesAddr,&checksum);
    if (NumBytesAddr == 0) {
      warning("Unknown 'S' type at line %i.\n",LineNumber);
      continue;
    }
    data.DataLength = data.count-NumBytesAddr-1;
    for (i = 0; i < data.DataLength; i++)
      checksum += data.data[i] = ReadByte(&p);
    data.checksum = ReadByte(&p);
    checksum = ~checksum;
    if (checksum != data.checksum) {
      warning("Checksum error at line %i.\n",LineNumber);
      continue;
    }

    /* process the valid data */
    proc_data(data);
  }
}

int main(int argc, char *argv[]) {
  int i;

  ProgramName=argv[0];

  SRecLoader();

  for (i=0;i<TPU_RAM_SIZE-1;i++) {
    printf("0x%02x,",tpu_buffer[i]);
    if (!((i+1) % 16)) putchar('\n');
  }
  printf("0x%02x\n",tpu_buffer[TPU_RAM_SIZE-1]);

  exit(0);
}

