#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <EFI.h>
#include "tpu.h"

#define EFI_TPU_IARB 13
#define EFI_TPU_CIRL 5
#define EFI_TPU_CIBV (96 >> 4)	/* Vectors 96-111 */
#define EFI_TPU0 96
#define EFI_TPU1 97
#define TPU_RAM_SIZE (2*1024)
#define TRAM_START 0xf0000

/* void SRecLoader(void); */
typedef unsigned char byte;

const byte tpu_buffer[TPU_RAM_SIZE] = {
#include "wheel.tpu"
};

const char *ProgramName="Wheel";
int angle = 0;

void warning(char *fmt, ...) {
  va_list args;
  
  va_start(args, fmt);
  fprintf(stderr,"%s: ", ProgramName);
  vfprintf(stderr, fmt, args);
  va_end(args);
}

extern void _catchTPU0int();
asm("    .text
         .align 2
         .globl _catchTPU0int
_catchTPU0int:
         moveml %d0-%d7/%a0-%a6,%sp@-
         jbsr   tpu0_interrupt
         moveml %sp@+,%d0-%d7/%a0-%a6
         rte
    ");

extern void _catchTPU1int();
asm("    .text
         .align 2
         .globl _catchTPU0int
_catchTPU1int:
         moveml %d0-%d7/%a0-%a6,%sp@-
         jbsr   tpu1_interrupt
         moveml %sp@+,%d0-%d7/%a0-%a6
         rte
    ");

void tpu0_interrupt(void) {
  warning("Received TPU0 interrupt. status=%x\n",*(TPU_PR0+2)>>8);
  *(TPU_PR0+2) = 0x7f;
  *CISR &= 0xfffe;
}

void tpu1_interrupt(void) {
  warning("Received TPU1 interrupt.\n");
  *CISR &= 0xfffd;
}

int main() {

  if (setvbuf(stdout,NULL,_IONBF,1)!=0) {
    perror("my error");
    exit(1);
  }

  /* Configure TRAM */

  *TRAMBAR = ((TRAM_START>>8) & TRAM_ADDR); /* Base address 0xf0000 */
  *TRAMMCR = 0;

  /* fill TRAM w/ TPU code */
  bcopy(tpu_buffer,(byte *)TRAM_START,TPU_RAM_SIZE);

  /* TPU interrput vector */

/*   PIT_HOOK = pit_hook; */
  exceptionHandler(EFI_TPU0,_catchTPU0int);
  exceptionHandler(EFI_TPU1,_catchTPU1int);

  /* TPU Initailization */

  *TPUMCR = ( SAM(0x0,13,TPU_TCR1P) | /* TCR1 Prescaler to 1 */
	      SAM(EFI_TPU_IARB,0,TPU_IARB) | /* Interrupt Arbitration */
	      TPU_EMU );

  *TICR = (SAM(EFI_TPU_CIRL,8,TPU_CIRL) | SAM(EFI_TPU_CIBV,4,TPU_CIBV));

  /* Channel Fuction Select Registers */
  *CFSR0 = 0;
  *CFSR1 = 0;
  *CFSR2 = 0;
  *CFSR3 = SAM(0xe,4,0x00f0) | SAM(0xf,0,0x000f);

  /* Parameters */
  *(TPU_PR0+0) = 0x4;		/* Detect rising edge */
  *(TPU_PR0+1) = 30;		/* WAIT_TEETH */
  *(TPU_PR0+2) = 0x7f;		/* STATUS_NUMBER/INT_TOOTH */
  *(TPU_PR0+3) = 0x0;		/* START_PERIOD/NUMBER_OF_PERIODS */
  *(TPU_PR0+4) = (96<<8);	/* RATIO/COUNT_360 */
  *(TPU_PR0+5) = (1<<12)+(1<<8); /* LINK_START/LINK_COUNT/COUNT_720 */

  *(TPU_PR1+0) = 0x8A;		/* CONFIG/EXTRA_TIME */
  *(TPU_PR1+1) = 5;		/* HIGH_TIME */
  *(TPU_PR1+2) = (15<<8);	/* START_ANGLE */
  *(TPU_PR1+3) = 0;		/*  */
  *(TPU_PR1+4) = 0;		/*  */
  *(TPU_PR1+5) = (60<<8)+((int)(TPU_PR0+4) & 0x7e); /* AD/SPACING/EX/COUNT_PTR/
					        PL */

  /* Host Sequence Registers */
  *HSQR1 = SAM(0x3,2,0xc);

  /* Host Service Request Register */
	/* Initalize */
  *HSRR1 = SAM(0x3,2,0xc) | SAM(0x3,0,0x3);
  
  /* Channel Interrup1 Register */
  *CIER = 1;

  /* Channel Priority Register */
  *CPR0 = 0;
  *CPR1 = SAM(TPU_Middle,2,0xc) | SAM(TPU_High,0,0x3);

  /* Wait for initalization to complete */
  while (*HSRR1 & 0xf);

  /* End TPU Initalization */

  for (;;) {
    scanf("%i",&angle);
    *(TPU_PR1+2) = (short int)(angle<<8); /* START_ANGLE */
  }
}

  
